import torch
from torch import nn
from torch.nn import functional as F
import math

from .conv import Conv2dTranspose, Conv2d, nonorm_Conv2d

class Wav2Lip(nn.Module):
    def __init__(self):
        super(Wav2Lip, self).__init__()

        self.face_encoder_blocks = nn.ModuleList([
            nn.Sequential(Conv2d(6, 16, kernel_size=7, stride=1, padding=3)), # 96,96

            nn.Sequential(Conv2d(16, 32, kernel_size=3, stride=2, padding=1), # 48,48
            Conv2d(32, 32, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(32, 32, kernel_size=3, stride=1, padding=1, residual=True)),

            nn.Sequential(Conv2d(32, 64, kernel_size=3, stride=2, padding=1),    # 24,24
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True)),

            nn.Sequential(Conv2d(64, 128, kernel_size=3, stride=2, padding=1),   # 12,12
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True)),

            nn.Sequential(Conv2d(128, 256, kernel_size=3, stride=2, padding=1),       # 6,6
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True)),

            nn.Sequential(Conv2d(256, 512, kernel_size=3, stride=2, padding=1),     # 3,3
            Conv2d(512, 512, kernel_size=3, stride=1, padding=1, residual=True),),
            
            nn.Sequential(Conv2d(512, 512, kernel_size=3, stride=1, padding=0),     # 1, 1
            Conv2d(512, 512, kernel_size=1, stride=1, padding=0)),])

        self.audio_encoder = nn.Sequential(
            Conv2d(1, 32, kernel_size=3, stride=1, padding=1),
            Conv2d(32, 32, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(32, 32, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(32, 64, kernel_size=3, stride=(3, 1), padding=1),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(64, 128, kernel_size=3, stride=3, padding=1),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(128, 256, kernel_size=3, stride=(3, 2), padding=1),
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(256, 512, kernel_size=3, stride=1, padding=0),
            Conv2d(512, 512, kernel_size=1, stride=1, padding=0),)

        self.face_decoder_blocks = nn.ModuleList([
            nn.Sequential(Conv2d(512, 512, kernel_size=1, stride=1, padding=0),),

            nn.Sequential(Conv2dTranspose(1024, 512, kernel_size=3, stride=1, padding=0), # 3,3
            Conv2d(512, 512, kernel_size=3, stride=1, padding=1, residual=True),),

            nn.Sequential(Conv2dTranspose(1024, 512, kernel_size=3, stride=2, padding=1, output_padding=1),
            Conv2d(512, 512, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(512, 512, kernel_size=3, stride=1, padding=1, residual=True),), # 6, 6

            nn.Sequential(Conv2dTranspose(768, 384, kernel_size=3, stride=2, padding=1, output_padding=1),
            Conv2d(384, 384, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(384, 384, kernel_size=3, stride=1, padding=1, residual=True),), # 12, 12

            nn.Sequential(Conv2dTranspose(512, 256, kernel_size=3, stride=2, padding=1, output_padding=1),
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True),), # 24, 24

            nn.Sequential(Conv2dTranspose(320, 128, kernel_size=3, stride=2, padding=1, output_padding=1), 
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),), # 48, 48

            nn.Sequential(Conv2dTranspose(160, 64, kernel_size=3, stride=2, padding=1, output_padding=1),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),),]) # 96,96

        self.output_block = nn.Sequential(Conv2d(80, 32, kernel_size=3, stride=1, padding=1),
            nn.Conv2d(32, 3, kernel_size=1, stride=1, padding=0),
            nn.Sigmoid()) 

    def forward(self, audio_sequences, face_sequences):
        # audio_sequences = (B, T, 1, 80, 16)
        B = audio_sequences.size(0)

        input_dim_size = len(face_sequences.size())
        if input_dim_size > 4:
            audio_sequences = torch.cat([audio_sequences[:, i] for i in range(audio_sequences.size(1))], dim=0)
            face_sequences = torch.cat([face_sequences[:, :, i] for i in range(face_sequences.size(2))], dim=0)

        audio_embedding = self.audio_encoder(audio_sequences) # B, 512, 1, 1

        feats = []
        x = face_sequences
        for f in self.face_encoder_blocks:
            x = f(x)
            feats.append(x)

        x = audio_embedding
        for f in self.face_decoder_blocks:
            x = f(x)
            try:
                x = torch.cat((x, feats[-1]), dim=1)
            except Exception as e:
                print(x.size())
                print(feats[-1].size())
                raise e
            
            feats.pop()

        x = self.output_block(x)

        if input_dim_size > 4:
            x = torch.split(x, B, dim=0) # [(B, C, H, W)]
            outputs = torch.stack(x, dim=2) # (B, C, T, H, W)

        else:
            outputs = x
            
        return outputs

class Wav2Lip_disc_qual(nn.Module):
    def __init__(self):
        super(Wav2Lip_disc_qual, self).__init__()

        self.face_encoder_blocks = nn.ModuleList([
            nn.Sequential(nonorm_Conv2d(3, 32, kernel_size=7, stride=1, padding=3)), # 48,96

            nn.Sequential(nonorm_Conv2d(32, 64, kernel_size=5, stride=(1, 2), padding=2), # 48,48
            nonorm_Conv2d(64, 64, kernel_size=5, stride=1, padding=2)),

            nn.Sequential(nonorm_Conv2d(64, 128, kernel_size=5, stride=2, padding=2),    # 24,24
            nonorm_Conv2d(128, 128, kernel_size=5, stride=1, padding=2)),

            nn.Sequential(nonorm_Conv2d(128, 256, kernel_size=5, stride=2, padding=2),   # 12,12
            nonorm_Conv2d(256, 256, kernel_size=5, stride=1, padding=2)),

            nn.Sequential(nonorm_Conv2d(256, 512, kernel_size=3, stride=2, padding=1),       # 6,6
            nonorm_Conv2d(512, 512, kernel_size=3, stride=1, padding=1)),

            nn.Sequential(nonorm_Conv2d(512, 512, kernel_size=3, stride=2, padding=1),     # 3,3
            nonorm_Conv2d(512, 512, kernel_size=3, stride=1, padding=1),),
            
            nn.Sequential(nonorm_Conv2d(512, 512, kernel_size=3, stride=1, padding=0),     # 1, 1
            nonorm_Conv2d(512, 512, kernel_size=1, stride=1, padding=0)),])

        self.binary_pred = nn.Sequential(nn.Conv2d(512, 1, kernel_size=1, stride=1, padding=0), nn.Sigmoid())
        self.label_noise = .0

    def get_lower_half(self, face_sequences):
        return face_sequences[:, :, face_sequences.size(2)//2:]

    def to_2d(self, face_sequences):
        B = face_sequences.size(0)
        face_sequences = torch.cat([face_sequences[:, :, i] for i in range(face_sequences.size(2))], dim=0)
        return face_sequences

    def perceptual_forward(self, false_face_sequences):
        false_face_sequences = self.to_2d(false_face_sequences)
        false_face_sequences = self.get_lower_half(false_face_sequences)

        false_feats = false_face_sequences
        for f in self.face_encoder_blocks:
            false_feats = f(false_feats)

        false_pred_loss = F.binary_cross_entropy(self.binary_pred(false_feats).view(len(false_feats), -1), 
                                        torch.ones((len(false_feats), 1)).cuda())

        return false_pred_loss

    def forward(self, face_sequences):
        face_sequences = self.to_2d(face_sequences)
        face_sequences = self.get_lower_half(face_sequences)

        x = face_sequences
        for f in self.face_encoder_blocks:
            x = f(x)

        return self.binary_pred(x).view(len(x), -1)



from glob import glob
import os

def get_image_list(data_root, split):
	filelist = []

	with open('filelists/{}.txt'.format(split)) as f:
		for line in f:
			line = line.strip()
			if ' ' in line: line = line.split()[0]
			filelist.append(os.path.join(data_root, line))

	return filelist

class HParams:
	def __init__(self, **kwargs):
		self.data = {}

		for key, value in kwargs.items():
			self.data[key] = value

	def __getattr__(self, key):
		if key not in self.data:
			raise AttributeError("'HParams' object has no attribute %s" % key)
		return self.data[key]

	def set_hparam(self, key, value):
		self.data[key] = value


# Default hyperparameters
hparams = HParams(
	num_mels=80,  # Number of mel-spectrogram channels and local conditioning dimensionality
	#  network
	rescale=True,  # Whether to rescale audio prior to preprocessing
	rescaling_max=0.9,  # Rescaling value
	
	# Use LWS (https://github.com/Jonathan-LeRoux/lws) for STFT and phase reconstruction
	# It"s preferred to set True to use with https://github.com/r9y9/wavenet_vocoder
	# Does not work if n_ffit is not multiple of hop_size!!
	use_lws=False,
	
	n_fft=800,  # Extra window size is filled with 0 paddings to match this parameter
	hop_size=200,  # For 16000Hz, 200 = 12.5 ms (0.0125 * sample_rate)
	win_size=800,  # For 16000Hz, 800 = 50 ms (If None, win_size = n_fft) (0.05 * sample_rate)
	sample_rate=16000,  # 16000Hz (corresponding to librispeech) (sox --i <filename>)
	
	frame_shift_ms=None,  # Can replace hop_size parameter. (Recommended: 12.5)
	
	# Mel and Linear spectrograms normalization/scaling and clipping
	signal_normalization=True,
	# Whether to normalize mel spectrograms to some predefined range (following below parameters)
	allow_clipping_in_normalization=True,  # Only relevant if mel_normalization = True
	symmetric_mels=True,
	# Whether to scale the data to be symmetric around 0. (Also multiplies the output range by 2, 
	# faster and cleaner convergence)
	max_abs_value=4.,
	# max absolute value of data. If symmetric, data will be [-max, max] else [0, max] (Must not 
	# be too big to avoid gradient explosion, 
	# not too small for fast convergence)
	# Contribution by @begeekmyfriend
	# Spectrogram Pre-Emphasis (Lfilter: Reduce spectrogram noise and helps model certitude 
	# levels. Also allows for better G&L phase reconstruction)
	preemphasize=True,  # whether to apply filter
	preemphasis=0.97,  # filter coefficient.
	
	# Limits
	min_level_db=-100,
	ref_level_db=20,
	fmin=55,
	# Set this to 55 if your speaker is male! if female, 95 should help taking off noise. (To 
	# test depending on dataset. Pitch info: male~[65, 260], female~[100, 525])
	fmax=7600,  # To be increased/reduced depending on data.

	###################### Our training parameters #################################
	img_size=96,
	fps=25,
	
	batch_size=16,
	initial_learning_rate=1e-4,
	nepochs=200000000000000000,  ### ctrl + c, stop whenever eval loss is consistently greater than train loss for ~10 epochs
	num_workers=16,
	checkpoint_interval=3000,
	eval_interval=3000,
    save_optimizer_state=True,

    syncnet_wt=0.0, # is initially zero, will be set automatically to 0.03 later. Leads to faster convergence. 
	syncnet_batch_size=64,
	syncnet_lr=1e-4,
	syncnet_eval_interval=10000,
	syncnet_checkpoint_interval=10000,

	disc_wt=0.07,
	disc_initial_learning_rate=1e-4,
)


def hparams_debug_string():
	values = hparams.values()
	hp = ["  %s: %s" % (name, values[name]) for name in sorted(values) if name != "sentences"]
	return "Hyperparameters:\n" + "\n".join(hp)


from os import listdir, path
import numpy as np
import scipy, cv2, os, sys, argparse
# import .video_generation.audio_mod.audio as audio
from .video_generation.audio_mod import audio as audio
import json, subprocess, random, string
from tqdm import tqdm
from glob import glob
import torch
from .video_generation import Wav2Lip
import time
import cv2



def run_wav2lip(checkpoint_path, metadata_path, audio_dir, output_dir='results/', wav2lip_batch_size=128, img_size=96):
    """
    Generate lip-sync videos using Wav2Lip models with preprocessed frames and face data
    Without attaching audio to the final video
    
    Parameters:
    checkpoint_path (str): Path to the saved checkpoint to load weights from
    metadata_path (str): Path to the metadata JSON file with frame paths and bounding boxes
    audio_dir (str): Directory containing audio files to process
    output_dir (str): Directory for output videos
    wav2lip_batch_size (int): Batch size for Wav2Lip model(s)
    img_size (int): Image size for the model
    
    Returns:
    dict: Summary of execution times
    """
    start_time = time.time()
    
    # Handle relative paths by converting to absolute paths
    current_dir = os.path.abspath(os.path.dirname(__file__))
    # checkpoint_path = os.path.abspath(os.path.join(current_dir, checkpoint_path))
    # metadata_path = os.path.abspath(os.path.join(current_dir, metadata_path))
    # audio_dir = os.path.abspath(os.path.join(current_dir, audio_dir))
    # output_dir = os.path.abspath(os.path.join(current_dir, output_dir))
    
    print(f"Using absolute paths:")
    print(f"Checkpoint: {checkpoint_path}")
    print(f"Metadata: {metadata_path}")
    print(f"Audio directory: {audio_dir}")
    print(f"Output directory: {output_dir}")
    
    mel_step_size = 16
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    print('Using {} for inference.'.format(device))
    
    # Create output directories
    os.makedirs('temp', exist_ok=True)
    os.makedirs(output_dir, exist_ok=True)
    
    def _load(checkpoint_path):
        if device == 'cuda':
            checkpoint = torch.load(checkpoint_path)
        else:
            checkpoint = torch.load(checkpoint_path,
                                    map_location=lambda storage, loc: storage)
        return checkpoint

    def load_model(path):
        model = Wav2Lip()
        print("Load checkpoint from: {}".format(path))
        checkpoint = _load(path)
        s = checkpoint["state_dict"]
        new_s = {}
        for k, v in s.items():
            new_s[k.replace('module.', '')] = v
        model.load_state_dict(new_s)

        model = model.to(device)
        return model.eval()

    def datagen(frames, mels, face_coords):
        img_batch, mel_batch, frame_batch, coords_batch = [], [], [], []

        for i, m in enumerate(mels):
            idx = i % len(frames)
            frame_to_save = frames[idx].copy()
            
            # Extract face using coordinates
            x1, y1, x2, y2 = face_coords[idx]
            face = frame_to_save[y1:y2, x1:x2].copy()

            face = cv2.resize(face, (img_size, img_size))
                
            img_batch.append(face)
            mel_batch.append(m)
            frame_batch.append(frame_to_save)
            coords_batch.append((y1, y2, x1, x2))

            if len(img_batch) >= wav2lip_batch_size:
                img_batch, mel_batch = np.asarray(img_batch), np.asarray(mel_batch)

                img_masked = img_batch.copy()
                img_masked[:, img_size//2:] = 0

                img_batch = np.concatenate((img_masked, img_batch), axis=3) / 255.
                mel_batch = np.reshape(mel_batch, [len(mel_batch), mel_batch.shape[1], mel_batch.shape[2], 1])

                yield img_batch, mel_batch, frame_batch, coords_batch
                img_batch, mel_batch, frame_batch, coords_batch = [], [], [], []

        if len(img_batch) > 0:
            img_batch, mel_batch = np.asarray(img_batch), np.asarray(mel_batch)

            img_masked = img_batch.copy()
            img_masked[:, img_size//2:] = 0

            img_batch = np.concatenate((img_masked, img_batch), axis=3) / 255.
            mel_batch = np.reshape(mel_batch, [len(mel_batch), mel_batch.shape[1], mel_batch.shape[2], 1])

            yield img_batch, mel_batch, frame_batch, coords_batch

    def process_audio_file(audio_path, frames, fps, face_coords, model):
        # Create an output path for the temp video (without audio)
        audio_filename = os.path.basename(audio_path).rsplit('.', 1)[0]
        # Save the final video to the output directory
        final_output_path = os.path.join(output_dir, f'{audio_filename}.mp4')
        
        # Time audio preprocessing
        audio_prep_start = time.time()
        if not audio_path.endswith('.wav'):
            print(f'Extracting raw audio from {audio_path}...')
            temp_dir = os.path.join(current_dir, 'temp')
            os.makedirs(temp_dir, exist_ok=True)
            temp_wav = os.path.join(temp_dir, f'temp_{audio_filename}.wav')
            command = f'ffmpeg -y -i "{audio_path}" -strict -2 "{temp_wav}"'
            subprocess.call(command, shell=True)
            audio_path = temp_wav

        wav = audio.load_wav(audio_path, 16000)
        mel = audio.melspectrogram(wav)
        audio_prep_time = time.time() - audio_prep_start
        print(f'Audio preprocessing completed in {audio_prep_time:.2f} seconds')
        
        if np.isnan(mel.reshape(-1)).sum() > 0:
            print(f'Warning: Mel contains NaN for {audio_path}. Skipping.')
            return

        # Time mel chunks preparation
        mel_prep_start = time.time()
        mel_chunks = []
        mel_idx_multiplier = 80./fps 
        i = 0
        while 1:
            start_idx = int(i * mel_idx_multiplier)
            if start_idx + mel_step_size > len(mel[0]):
                mel_chunks.append(mel[:, len(mel[0]) - mel_step_size:])
                break
            mel_chunks.append(mel[:, start_idx : start_idx + mel_step_size])
            i += 1
        mel_prep_time = time.time() - mel_prep_start
        print(f'Mel chunks preparation completed in {mel_prep_time:.2f} seconds')

        print(f"Length of mel chunks for {os.path.basename(audio_path)}: {len(mel_chunks)}")
        
        num_frames_needed = len(mel_chunks)
        num_frames_available = len(frames)
        
        print(f"Audio requires {num_frames_needed} frames, video has {num_frames_available} frames")
        
        frame_h, frame_w = frames[0].shape[:-1]
        temp_dir = os.path.join(current_dir, 'temp')
        os.makedirs(temp_dir, exist_ok=True)
        
        # Save the temp video with the audio filename instead of a generic name
        temp_video_path = os.path.join(temp_dir, f'{audio_filename}.avi')
        out = cv2.VideoWriter(temp_video_path, 
                             cv2.VideoWriter_fourcc(*'DIVX'), fps, (frame_w, frame_h))

        # Time lip generation
        lip_gen_start = time.time()
        batch_size = wav2lip_batch_size
        gen = datagen(frames.copy(), mel_chunks, face_coords)

        for i, (img_batch, mel_batch, frames_batch, coords) in enumerate(tqdm(gen, 
                                                total=int(np.ceil(float(len(mel_chunks))/batch_size)),
                                                desc=f"Processing {os.path.basename(audio_path)}")):
            img_batch = torch.FloatTensor(np.transpose(img_batch, (0, 3, 1, 2))).to(device)
            mel_batch = torch.FloatTensor(np.transpose(mel_batch, (0, 3, 1, 2))).to(device)

            with torch.no_grad():
                pred = model(mel_batch, img_batch)

            pred = pred.cpu().numpy().transpose(0, 2, 3, 1) * 255.
            
            for p, f, c in zip(pred, frames_batch, coords):
                y1, y2, x1, x2 = c
                p = cv2.resize(p.astype(np.uint8), (x2 - x1, y2 - y1))

                f[y1:y2, x1:x2] = p
                out.write(f)

        out.release()
        lip_gen_time = time.time() - lip_gen_start
        print(f'Lip generation completed in {lip_gen_time:.2f} seconds')
        print("the temp video generated path is", temp_video_path)
        # convert_avi_to_mp4(temp_video_path, final_output_path)
        # Instead of combining with audio, just copy the temp video to the output directory
        # video_creation_start = time.time()
        
        # Just copy the temp video file to the output directory
        # command = f'ffmpeg -y -i "{temp_video_path}" -c:v copy "{final_output_path}"'
        # subprocess.call(command, shell=platform.system() != 'Windows')
        
        # video_creation_time = time.time() - video_creation_start
        # print(f'Video file copied to output directory in {video_creation_time:.2f} seconds')
        
        # Don't remove the temp file as that's what we want to keep
        # print(f'Temporary video file saved at: {temp_video_path}')
        # print(f'Final video file (without audio) saved at: {final_output_path}')
        video_creation_time = 0
        return {
            'audio_prep_time': audio_prep_time,
            'mel_prep_time': mel_prep_time,
            'lip_gen_time': lip_gen_time,
            # 'video_creation_time': video_creation_time,
            'total_time': audio_prep_time + mel_prep_time + lip_gen_time + video_creation_time,
            'temp_video_path': temp_video_path,
            'final_output_path': final_output_path
        }

    # Load metadata
    print(f"Loading metadata from {metadata_path}")
    with open(metadata_path, 'r') as f:
        metadata = json.load(f)
    
    fps = metadata['fps']
    print(f"Video FPS: {fps}")
    
    # Load all frames
    print("Loading frames...")
    frames = []
    face_coords = []
    
    for frame_data in tqdm(metadata['frames'], desc="Loading frames"):
        frame_path = frame_data['frame_path']
        # print(frame_path)
        new_path = frame_path.replace("output\\", "")
        # print(new_path)
        # Convert relative frame path to absolute if needed
        if not os.path.isabs(new_path):
            # Get directory of metadata file to use as base for relative frame paths
            metadata_dir = os.path.dirname(metadata_path)
            new_path = os.path.abspath(os.path.join((metadata_dir), new_path))
        
        bbox = frame_data['bbox']
        
        frame = cv2.imread(new_path)
        if frame is None:
            raise ValueError(f"Could not load frame: {new_path}")
        
        frames.append(frame)
        face_coords.append(bbox)
    
    print(f"Loaded {len(frames)} frames")
    
    # Time model loading
    model_loading_start = time.time()
    model = load_model(checkpoint_path)
    model_loading_time = time.time() - model_loading_start
    print(f'Model loading completed in {model_loading_time:.2f} seconds')
    
    # Process audio files
    if os.path.isdir(audio_dir):
        audio_files = []
        for ext in ['wav', 'mp3', 'm4a', 'aac', 'flac']:
            audio_files.extend(glob(os.path.join(audio_dir, f'*.{ext}')))
        
        if not audio_files:
            raise ValueError(f"No audio files found in {audio_dir}")
        
        print(f"Found {len(audio_files)} audio files to process")
        video_generation_start = time.time()
        
        audio_processing_results = []
        for audio_file in audio_files:
            each_audio_start = time.time() 
            print(f"\nProcessing {os.path.basename(audio_file)}...")
            result = process_audio_file(audio_file, frames, fps, face_coords, model)
            if result:
                audio_processing_results.append({
                    'audio_file': audio_file,
                    'processing_time': time.time() - each_audio_start,
                    'details': result
                })
            print(f"Audio file {os.path.basename(audio_file)} processed in {time.time() - each_audio_start:.2f} seconds")
        
        video_generation_time = time.time() - video_generation_start
        print(f"Total video generation time: {video_generation_time:.2f} seconds")
            
    else:
        raise ValueError('audio_dir must be a valid directory containing audio files')
    
    total_time = time.time() - start_time
    execution_summary = {
        'model_loading_time': model_loading_time,
        'video_generation_time': video_generation_time,
        'total_execution_time': total_time,
        'audio_processing_results': audio_processing_results
    }
    
    print(f"\nTotal execution summary:")
    print(f"Model loading time: {model_loading_time:.2f}s")
    print(f"Video generation time: {video_generation_time:.2f}s")
    print(f"Total execution time: {total_time:.2f}s")
    
    return execution_summary



# Example of how to use this function directly in another script:
if __name__ == '__main__':

    run_wav2lip(
        checkpoint_path="checkpoints/wav2lip.pth",
        metadata_path="../output/metadata.json",
        audio_dir= "../input_audio/",
        output_dir="../output_videos/",
    )


from os import listdir, path
import numpy as np
import cv2
import os
import json
from tqdm import tqdm
import torch
from .face_detection import api as face_detect_w
import time
from pathlib import Path

def extract_faces(
    face_path,
    output_dir,
    static=False,
    fps=25.0,
    pads=[0, 10, 0, 0],
    face_det_batch_size=16,
    resize_factor=2,
    crop=[0, -1, 0, -1],
    box=[-1, -1, -1, -1],
    rotate=False,
    nosmooth=False
):
    """
    Extract frames and face bounding boxes from video for Wav2Lip
    
    Parameters:
    face_path (str): Filepath of video/image that contains faces to use
    output_dir (str): Directory to save frames and face data
    static (bool): If True, then use only first video frame for inference
    fps (float): Can be specified only if input is a static image
    pads (list): Padding (top, bottom, left, right). Please adjust to include chin at least
    face_det_batch_size (int): Batch size for face detection
    resize_factor (int): Reduce the resolution by this factor
    crop (list): Crop video to a smaller region (top, bottom, left, right)
    box (list): Specify a constant bounding box for the face
    rotate (bool): If true, will flip video right by 90deg
    nosmooth (bool): Prevent smoothing face detections over a short temporal window
    
    Returns:
    dict: Metadata about the processed frames and face boxes
    """
    start_time = time.time()
    print("batch size is ", face_det_batch_size)
    img_size = 96
    # face_path = os.path.normpath(face_path)
    print(face_path)
    if not os.path.isfile(face_path):
        raise ValueError('face_path argument must be a valid path to video/image file')
    
    if os.path.isfile(face_path) and face_path.split('.')[1] in ['jpg', 'png', 'jpeg']:
        static = True
    
    # Create output directories
    frames_dir = os.path.join(output_dir, 'frames')
    os.makedirs(frames_dir, exist_ok=True)
    
    print(f"Using device: {'cuda' if torch.cuda.is_available() else 'cpu'}")
    
    # Time video frame reading and saving
    print('Reading and saving video frames...')
    full_frames = []
    frame_paths = []
    
    if face_path.split('.')[1] in ['jpg', 'png', 'jpeg']:
        frame = cv2.imread(face_path)
        
        # Apply resize and crop if needed
        if resize_factor > 1:
            frame = cv2.resize(frame, (frame.shape[1]//resize_factor, frame.shape[0]//resize_factor))
        
        if rotate:
            frame = cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
        
        y1, y2, x1, x2 = crop
        if x2 == -1: x2 = frame.shape[1]
        if y2 == -1: y2 = frame.shape[0]
        
        frame = frame[y1:y2, x1:x2]
        full_frames.append(frame)
        
        # Save the frame
        frame_path = os.path.join(frames_dir, f"frame_0.jpg")
        cv2.imwrite(frame_path, frame)
        frame_paths.append(frame_path)
        
    else:
        video_stream = cv2.VideoCapture(face_path)
        fps = video_stream.get(cv2.CAP_PROP_FPS)
        
        frame_idx = 0
        while True:
            still_reading, frame = video_stream.read()
            if not still_reading:
                video_stream.release()
                break
                
            if resize_factor > 1:
                frame = cv2.resize(frame, (frame.shape[1]//resize_factor, frame.shape[0]//resize_factor))
            
            if rotate:
                frame = cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
            
            y1, y2, x1, x2 = crop
            if x2 == -1: x2 = frame.shape[1]
            if y2 == -1: y2 = frame.shape[0]
            
            frame = frame[y1:y2, x1:x2]
            full_frames.append(frame)
            
            # Save the frame
            frame_path = os.path.join(frames_dir, f"frame_{frame_idx}.jpg")
            cv2.imwrite(frame_path, frame)
            frame_paths.append(frame_path)
            frame_idx += 1
    
    print(f"Total frames extracted and saved: {len(full_frames)}")
    
    # Helper functions
    def get_smoothened_boxes(boxes, T):
        for i in range(len(boxes)):
            if i + T > len(boxes):
                window = boxes[len(boxes) - T:]
            else:
                window = boxes[i : i + T]
            boxes[i] = np.mean(window, axis=0)
        return boxes

    def face_detect(images):
        start_time = time.time()
        
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        detector = face_detect_w.FaceAlignment(face_detect_w.LandmarksType._2D, 
                                                flip_input=False, device=device)

        batch_size = face_det_batch_size
        print("batch size(2)", face_det_batch_size)
        
        while 1:
            predictions = []
            try:
                for i in tqdm(range(0, len(images), batch_size), desc="Face detection"):
                    predictions.extend(detector.get_detections_for_batch(np.array(images[i:i + batch_size])))
            except RuntimeError as e:
                print(batch_size)
                print(e)
                # if batch_size == 1: 
                #     raise RuntimeError('Image too big to run face detection on GPU. Please use the resize_factor parameter')
                if batch_size != 1:
                    batch_size //= 2
                    print('Recovering from OOM error; New batch size: {}'.format(batch_size))
                    continue
            break

        results = []
        pady1, pady2, padx1, padx2 = pads
        for rect, image in zip(predictions, images):
            if rect is None:
                cv2.imwrite(os.path.join(output_dir, 'faulty_frame.jpg'), image)
                raise ValueError('Face not detected! Ensure the video contains a face in all the frames.')

            y1 = max(0, rect[1] - pady1)
            y2 = min(image.shape[0], rect[3] + pady2)
            x1 = max(0, rect[0] - padx1)
            x2 = min(image.shape[1], rect[2] + padx2)
            
            results.append([x1, y1, x2, y2])

        boxes = np.array(results)
        if not nosmooth: 
            boxes = get_smoothened_boxes(boxes, T=5)
        
        face_detect_time = time.time() - start_time
        print(f'Face detection completed in {face_detect_time:.2f} seconds')
        
        return boxes
    
    # Detect faces or use specified box
    if box[0] == -1:
        print("Running face detection...")
        if static:
            face_boxes = face_detect([full_frames[0]])
            # Replicate the same box for all frames
            face_boxes = np.tile(face_boxes, (len(full_frames), 1))
        else:
            face_boxes = face_detect(full_frames)
    else:
        print('Using the specified bounding box instead of face detection...')
        y1, y2, x1, x2 = box
        # Create bounding boxes for all frames using the specified coordinates
        face_boxes = np.array([[x1, y1, x2, y2]] * len(full_frames))
    
    # Save face bounding box data with corresponding frame paths
    data = {
        'fps': fps,
        'frames': []
    }
    
    for i, (frame_path, box) in enumerate(zip(frame_paths, face_boxes)):
        data['frames'].append({
            'frame_path': frame_path,
            'bbox': box.tolist()  # [x1, y1, x2, y2]
        })
    
    # Save metadata to JSON file
    metadata_path = os.path.join(output_dir, 'metadata.json')
    with open(metadata_path, 'w') as f:
        json.dump(data, f, indent=4)
    
    print(f"Metadata saved to: {metadata_path}")
    print(f"Frames saved to: {frames_dir}")
    
    total_time = time.time() - start_time
    print(f"Total execution time: {total_time:.2f} seconds")
    
    return data

# Example usage
# if __name__ == '__main__':
#     # Example of how to call the function
#     result = extract_faces(
#         face_path= "../inputs/video.mp4",
#         output_dir='output',
#         resize_factor=2,
#         pads=[0, 10, 0, 0],
#     )


from setuptools import setup, find_packages

# Read requirements from requirements.txt
with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='lip_generation',
    version='0.4.0',
    packages=find_packages(),
    install_requires=required,
)


from __future__ import print_function
import os
import torch
from torch.utils.model_zoo import load_url
from enum import Enum
import numpy as np
import cv2
try:
    import urllib.request as request_file
except BaseException:
    import urllib as request_file

from .models import FAN, ResNetDepth
from .utils import *
from .detection.sfd import FaceDetector


class LandmarksType(Enum):
    """Enum class defining the type of landmarks to detect.

    ``_2D`` - the detected points ``(x,y)`` are detected in a 2D space and follow the visible contour of the face
    ``_2halfD`` - this points represent the projection of the 3D points into 3D
    ``_3D`` - detect the points ``(x,y,z)``` in a 3D space

    """
    _2D = 1
    _2halfD = 2
    _3D = 3


class NetworkSize(Enum):
    # TINY = 1
    # SMALL = 2
    # MEDIUM = 3
    LARGE = 4

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value

ROOT = os.path.dirname(os.path.abspath(__file__))

class FaceAlignment:
    def __init__(self, landmarks_type, network_size=NetworkSize.LARGE,
                 device='cuda', flip_input=False, face_detector='sfd', verbose=False):
        self.device = device
        self.flip_input = flip_input
        self.landmarks_type = landmarks_type
        self.verbose = verbose

        network_size = int(network_size)

        if 'cuda' in device:
            torch.backends.cudnn.benchmark = True

        # Get the face detector
        self.face_detector = FaceDetector(device= device, verbose= verbose)
        # face_detector_module = __import__('face_detection.detection.' + face_detector,
        #                                   globals(), locals(), [face_detector], 0)
        # self.face_detector = face_detector_module.FaceDetector(device=device, verbose=verbose)

    def get_detections_for_batch(self, images):
        print("hi")
        images = images[..., ::-1]
        detected_faces = self.face_detector.detect_from_batch(images.copy())
        print("double check")
        results = []

        for i, d in enumerate(detected_faces):
            if len(d) == 0:
                results.append(None)
                continue
            d = d[0]
            d = np.clip(d, 0, None)
            
            x1, y1, x2, y2 = map(int, d[:-1])
            results.append((x1, y1, x2, y2))

        return results